from typing import Any


def assert_dict(actual: dict, expected: dict):
    _assert_recursive_equals(path=[], actual=actual, expected=expected)


def _assert_recursive_equals(path: list[str], actual: Any, expected: Any):
    if isinstance(expected, dict) and isinstance(actual, dict):
        for key, value in expected.items():
            _assert_recursive_equals(path=path + [key], actual=actual.get(key), expected=value)
    elif isinstance(expected, list) and isinstance(actual, list):
        for index, element in enumerate(expected):
            superset_value = actual[index] if len(actual) > index else None
            _assert_recursive_equals(path=path + [str(index)], actual=superset_value, expected=element)
    elif callable(expected):
        expected(actual)
    else:
        assert expected == actual, f"{'.'.join(path)} expected {expected}, but was {actual}"


def matcher_string_contains(expected: str):
    def check(x: str):
        assert isinstance(x, str), f"Expected a string, but was {type(x)}, value={x}"
        assert expected in x, f"Expected '{x}' to contain '{expected}', value={x}"
        return x

    return check



from __future__ import annotations

import os
from typing import Optional

import pytest
from dotenv import load_dotenv
from helpers.data_source_test_helper import DataSourceTestHelper
from soda_core.common.logs import Logs

project_root_dir = __file__[: -len("/soda-core/tests/helpers/test_fixtures.py")]
load_dotenv(f"{project_root_dir}/.env", override=True)


# In global scope because it is used in pytest annotations, it would not work as a fixture.
test_datasource = os.getenv("TEST_DATASOURCE", "postgres")

in_memory_data_sources = [
    "duckdb",
]


def is_in_memory_data_source() -> bool:
    return test_datasource in in_memory_data_sources


def is_sampling_supported_data_source() -> bool:
    supported_sampling_data_sources = [
        "snowflake",
    ]
    return test_datasource in supported_sampling_data_sources


@pytest.fixture(scope="function")
def env_vars() -> dict:
    original_env_vars = dict(os.environ)
    yield os.environ
    os.environ.clear()
    os.environ.update(original_env_vars)


@pytest.fixture(scope="function")
def data_source_test_helper(data_source_test_helper_session: DataSourceTestHelper) -> DataSourceTestHelper:
    yield data_source_test_helper_session
    data_source_test_helper_session.test_method_ended()


@pytest.fixture(scope="function")
def secondary_data_source_test_helper(
    secondary_data_source_test_helper_session: DataSourceTestHelper,
) -> DataSourceTestHelper:
    yield secondary_data_source_test_helper_session
    secondary_data_source_test_helper_session.test_method_ended()


@pytest.fixture(scope="function")
def logs() -> Logs:
    logs: Logs = Logs()
    try:
        yield logs
    finally:
        logs.remove_from_root_logger()


@pytest.fixture(scope="session")
def data_source_test_helper_session() -> DataSourceTestHelper:
    data_source_test_helper: DataSourceTestHelper = DataSourceTestHelper.create(
        test_datasource,
        name="primary_datasource",  # Do not use '-' in the name, it will cause issues with the dwh when creating tables: some data sources (e.g. Athena) do not support hyphens in table names.
    )
    data_source_test_helper.start_test_session()
    exception: Optional[Exception] = None
    try:
        yield data_source_test_helper
    except Exception as e:
        exception = e
    finally:
        data_source_test_helper.end_test_session(exception=exception)


@pytest.fixture(scope="session")
def secondary_data_source_test_helper_session() -> DataSourceTestHelper:
    data_source_test_helper: DataSourceTestHelper = DataSourceTestHelper.create(
        test_datasource, name="secondary_datasource"
    )
    data_source_test_helper.start_test_session()
    exception: Optional[Exception] = None
    try:
        yield data_source_test_helper
    except Exception as e:
        exception = e
    finally:
        data_source_test_helper.end_test_session(exception=exception)



from __future__ import annotations

import datetime
import logging
import os
import random
import re
import string
from copy import deepcopy
from textwrap import dedent
from typing import Optional

from helpers.mock_soda_cloud import MockResponse, MockSodaCloud
from helpers.test_table import TestColumn, TestTable, TestTableSpecification
from soda_core.common.data_source_impl import DataSourceImpl
from soda_core.common.logs import Logs
from soda_core.common.metadata_types import SodaDataTypeName, SqlDataType
from soda_core.common.soda_cloud import SodaCloud
from soda_core.common.sql_ast import (
    COLUMN,
    CREATE_MATERIALIZED_VIEW,
    CREATE_TABLE,
    CREATE_TABLE_COLUMN,
    CREATE_VIEW,
    DROP_MATERIALIZED_VIEW,
    DROP_TABLE,
    DROP_VIEW,
    FROM,
    INSERT_INTO,
    SELECT,
    STAR,
    VALUES_ROW,
)
from soda_core.common.statements.metadata_tables_query import (
    FullyQualifiedTableName,
    FullyQualifiedViewName,
    MetadataTablesQuery,
    TableType,
)
from soda_core.common.statements.table_types import FullyQualifiedMaterializedViewName
from soda_core.common.yaml import (
    ContractYamlSource,
    DataSourceYamlSource,
    SodaCloudYamlSource,
)
from soda_core.contracts.contract_verification import (
    ContractVerificationResult,
    ContractVerificationSession,
    ContractVerificationSessionResult,
)

logger = logging.getLogger(__name__)


class DataSourceTestHelper:
    @classmethod
    def create(cls, test_datasource: str, name: str) -> DataSourceTestHelper:
        if test_datasource == "postgres":
            from soda_postgres.test_helpers.postgres_data_source_test_helper import (
                PostgresDataSourceTestHelper,
            )

            return PostgresDataSourceTestHelper(name)
        elif test_datasource == "snowflake":
            from soda_snowflake.test_helpers.snowflake_data_source_test_helper import (
                SnowflakeDataSourceTestHelper,
            )

            return SnowflakeDataSourceTestHelper(name)
        elif test_datasource == "databricks":
            from soda_databricks.test_helpers.databricks_data_source_test_helper import (
                DatabricksDataSourceTestHelper,
            )

            return DatabricksDataSourceTestHelper(name)
        elif test_datasource == "duckdb":
            from soda_duckdb.test_helpers.duckdb_data_source_test_helper import (
                DuckdbDataSourceTestHelper,
            )

            return DuckdbDataSourceTestHelper(name)
        elif test_datasource == "bigquery":
            from soda_bigquery.test_helpers.bigquery_data_source_test_helper import (
                BigQueryDataSourceTestHelper,
            )

            return BigQueryDataSourceTestHelper(name)

        elif test_datasource == "oracle":
            from soda_oracle.test_helpers.oracle_data_source_test_helper import (
                OracleDataSourceTestHelper,
            )

            return OracleDataSourceTestHelper(name)

        elif test_datasource == "sqlserver":
            from soda_sqlserver.test_helpers.sqlserver_data_source_test_helper import (
                SqlServerDataSourceTestHelper,
            )

            return SqlServerDataSourceTestHelper(name)
        elif test_datasource == "synapse":
            from soda_synapse.test_helpers.synapse_data_source_test_helper import (
                SynapseDataSourceTestHelper,
            )

            return SynapseDataSourceTestHelper(name)
        elif test_datasource == "redshift":
            from soda_redshift.test_helpers.redshift_data_source_test_helper import (
                RedshiftDataSourceTestHelper,
            )

            return RedshiftDataSourceTestHelper(name)
        elif test_datasource == "fabric":
            from soda_fabric.test_helpers.fabric_data_source_test_helper import (
                FabricDataSourceTestHelper,
            )

            return FabricDataSourceTestHelper(name)
        elif test_datasource == "athena":
            from soda_athena.test_helpers.athena_data_source_test_helper import (
                AthenaDataSourceTestHelper,
            )

            return AthenaDataSourceTestHelper(name)
        elif test_datasource == "sparkdf":
            from soda_sparkdf.test_helpers.sparkdf_data_source_test_helper import (
                SparkDataFrameDataSourceTestHelper,
            )

            return SparkDataFrameDataSourceTestHelper(name)

        elif test_datasource == "dremio":
            from soda_dremio.test_helpers.dremio_data_source_test_helper import (
                DremioDataSourceTestHelper,
            )

            return DremioDataSourceTestHelper(name)
        else:
            raise AssertionError(f"Unknown test data source {test_datasource}")

    def __init__(self, name: str):
        self.name = name
        self.dataset_prefix: list[str] = self._create_dataset_prefix()
        logs: Logs = Logs()
        self.data_source_impl: "DataSourceImpl" = self._create_data_source_impl()
        logs.remove_from_root_logger()
        if logs.has_errors:
            raise RuntimeError(f"Couldn't create DataSource: {logs}")
        self.is_cicd = os.getenv("GITHUB_ACTIONS") is not None

        # Test table names that are present in the data source.
        # None means the data source is not queried
        self.existing_test_table_names: Optional[list[str]] = None

        # Maps TestTable to their unique_name property
        # (that is the the full table name composed of "SODATEST_" prefix, table purpose & test table hash)
        self.test_tables: dict[str, TestTable] = {}

        self.soda_cloud: Optional[SodaCloud] = None
        self.use_agent: bool = False

        if os.environ.get("SEND_RESULTS_TO_SODA_CLOUD") == "on":
            self.enable_soda_cloud()

    def enable_soda_cloud(self):
        logs: Logs = Logs()
        soda_cloud_yaml_str: str = """
            soda_cloud:
              host: ${env.SODA_CLOUD_HOST}
              api_key_id: ${env.SODA_CLOUD_API_KEY_ID}
              api_key_secret: ${env.SODA_CLOUD_API_KEY_SECRET}
        """
        soda_cloud_yaml_source: SodaCloudYamlSource = SodaCloudYamlSource.from_str(yaml_str=soda_cloud_yaml_str)
        self.soda_cloud = SodaCloud.from_yaml_source(
            soda_cloud_yaml_source=soda_cloud_yaml_source, provided_variable_values={}
        )
        if logs.has_errors:
            raise AssertionError(str(logs))

    def enable_soda_cloud_mock(self, responses: Optional[list[MockResponse]] = None):
        self.soda_cloud = MockSodaCloud(responses)

    def _create_data_source_impl(self) -> "DataSourceImpl":
        """
        Called in constructor to initialized self.data_source
        """
        logs: Logs = Logs()
        data_source_yaml_source: DataSourceYamlSource = self._create_data_source_yaml_source()
        from soda_core.common.data_source_impl import DataSourceImpl

        data_source_impl: DataSourceImpl = DataSourceImpl.from_yaml_source(data_source_yaml_source)
        assert not logs.has_errors
        return data_source_impl

    def _create_data_source_yaml_str(self) -> str:
        """
        Called in _create_data_source_impl to initialized self.data_source_impl
        self.database_name and self.schema_name are available if appropriate for the data source type
        """
        return ""

    def _create_data_source_yaml_source(self) -> DataSourceYamlSource:
        test_data_source_yaml_str: str = self._create_data_source_yaml_str()
        test_data_source_yaml_str = dedent(test_data_source_yaml_str).strip()
        return DataSourceYamlSource.from_str(yaml_str=test_data_source_yaml_str)

    def _create_dataset_prefix(self) -> list[str]:
        database_name: str = self._create_database_name()
        schema_name: str = self._create_schema_name()
        return [database_name, schema_name]

    def _create_database_name(self) -> Optional[str]:
        """
        Called in constructor to initialized self.database_name
        """
        return "soda_test"

    def _create_schema_name(self) -> Optional[str]:
        """
        Called in constructor to initialized self.schema_name
        """

        schema_name_parts = []

        github_ref_name = os.getenv("GITHUB_REF_NAME")
        github_head_ref = os.getenv("GITHUB_HEAD_REF")

        if not github_ref_name and not github_head_ref:
            user = os.getenv("USER", "anonymous")
            schema_name_parts.append("dev")
            schema_name_parts.append(user)

        else:
            python_version = os.getenv("PYTHON_VERSION")
            python_version_short = f'P{python_version.replace(".", "")}' if python_version else ""
            timestamp = datetime.datetime.now().strftime("%Y_%m_%dT%H_%M_%S_%f")

            def generate_random_alpha_num_str(length: int) -> str:
                return "".join(random.choice(string.ascii_lowercase + string.digits) for _ in range(length))

            if github_head_ref:
                github_head_ref_short = (
                    github_head_ref[:15] if github_head_ref and len(github_head_ref) > 15 else github_head_ref
                )
                schema_name_parts.append("ci")
                schema_name_parts.append(github_head_ref_short)
                schema_name_parts.append(python_version_short)
                schema_name_parts.append(timestamp)

            else:
                schema_name_parts.append("ci_main")
                schema_name_parts.append(python_version_short)
                schema_name_parts.append(timestamp)

        schema_name_raw = "_".join(schema_name_parts)
        schema_name = re.sub("[^0-9a-zA-Z]+", "_", schema_name_raw).lower()
        return schema_name

    def _adjust_schema_name(self, schema_name: str) -> str:
        return schema_name

    def start_test_session(self) -> None:
        self.start_test_session_open_connection()
        self.start_test_session_ensure_schema()

    def start_test_session_open_connection(self) -> None:
        logs: Logs = Logs()
        self.data_source_impl.open_connection()
        logs.remove_from_root_logger()
        if logs.has_errors:
            raise AssertionError(f"Connection creation has errors. See logs.")

    def start_test_session_ensure_schema(self) -> None:
        if self.is_cicd:
            self.drop_test_schema_if_exists()
        self.create_test_schema_if_not_exists()

    def end_test_session(self, exception: Optional[Exception]) -> None:
        self.end_test_session_drop_schema()
        self.end_test_session_close_connection()

    def end_test_session_close_connection(self) -> None:
        self.data_source_impl.close_connection()

    def end_test_session_drop_schema(self) -> None:
        if self.is_cicd:
            self.drop_test_schema_if_exists()

    def query_existing_test_tables(self) -> list[FullyQualifiedTableName]:
        database: Optional[str] = self.extract_database_from_prefix()
        schema: Optional[str] = self.extract_schema_from_prefix()

        metadata_tables_query: MetadataTablesQuery = self.data_source_impl.create_metadata_tables_query()
        fully_qualified_table_names: list[FullyQualifiedTableName] = metadata_tables_query.execute(
            database_name=database,
            schema_name=schema,
            include_table_name_like_filters=["SODATEST_%"],
        )
        return fully_qualified_table_names

    def query_existing_test_table_names(self) -> list[str]:
        fully_qualified_table_names = self.query_existing_test_tables()
        return [
            fully_qualified_test_table_name.table_name
            for fully_qualified_test_table_name in fully_qualified_table_names
            if isinstance(fully_qualified_test_table_name, FullyQualifiedTableName)
        ]

    def create_test_schema_if_not_exists(self) -> None:
        sql: str = self.create_test_schema_if_not_exists_sql()
        self.data_source_impl.execute_update(sql)
        post_create_sql = self.post_test_schema_create_sql()
        if post_create_sql:
            [self.data_source_impl.execute_update(cmd) for cmd in post_create_sql]

    def create_test_schema_if_not_exists_sql(self) -> str:
        return self.data_source_impl.sql_dialect.create_schema_if_not_exists_sql(self.dataset_prefix)

    def post_test_schema_create_sql(self) -> str:
        return self.data_source_impl.sql_dialect.post_schema_create_sql(self.dataset_prefix)

    def extract_database_from_prefix(self) -> Optional[str]:
        return self.data_source_impl.extract_database_from_prefix(self.dataset_prefix)

    def extract_schema_from_prefix(self) -> Optional[str]:
        return self.data_source_impl.extract_schema_from_prefix(self.dataset_prefix)

    def drop_test_schema_if_exists(self) -> None:
        schema = self.extract_schema_from_prefix()
        if not schema:
            raise AssertionError("Data source does not support schemas")

        self.drop_schema_if_exists(schema)

    def drop_schema_if_exists(self, schema: str) -> None:
        try:
            sql: str = self.drop_schema_if_exists_sql(schema=schema)
            self.data_source_impl.execute_update(sql)
        except Exception as e:
            logger.warning(f"Error dropping test schema: {e}")

    def drop_schema_if_exists_sql(self, schema: str) -> str:
        return f"DROP SCHEMA IF EXISTS {self.data_source_impl.sql_dialect.quote_default(schema)} CASCADE;"

    def drop_schema_sql(self, schema: str) -> str:
        return f"DROP SCHEMA {self.data_source_impl.sql_dialect.quote_default(schema)} CASCADE;"

    def ensure_test_table(
        self, test_table_specification: TestTableSpecification, force_recreate: bool = False
    ) -> TestTable:
        """
        Returns a test table with the given table data
        """
        if self.existing_test_table_names is None:
            self.existing_test_table_names = self.query_existing_test_table_names()

        # Specifically for BigQuery; when developing locally, the metadata might lag behind from the actual state of the tables.
        # Uncomment this to force the test table to be recreated every time.
        # self.existing_test_table_names = []

        test_table: TestTable = self.test_tables.get(test_table_specification.unique_name)
        if not test_table:
            test_table = self._create_test_table_python_object(test_table_specification)

            existing_test_table_names_lower: list[str] = [
                existing_test_table_name.lower() for existing_test_table_name in self.existing_test_table_names
            ]
            if (
                test_table_specification.unique_name.lower() not in existing_test_table_names_lower
                or not self.verify_test_table_row_count(test_table_specification)
                or force_recreate
            ):
                obsolete_table_names = [
                    existing_test_table
                    for existing_test_table in self.existing_test_table_names
                    if existing_test_table.lower().startswith(f"sodatest_{test_table_specification.name.lower()}_")
                ]
                if obsolete_table_names:
                    for obsolete_table_name in obsolete_table_names:
                        logger.debug(f"Test table {obsolete_table_name} has changed and will be recreated")
                        self._drop_test_table(table_name=obsolete_table_name)
                        self.existing_test_table_names.remove(obsolete_table_name)

                logger.debug(f"Test table {test_table_specification.unique_name} will be created")
                self._create_and_insert_test_table(test_table=test_table)
                self.existing_test_table_names.append(test_table.unique_name)

                self.data_source_impl.data_source_connection.commit()
        else:
            logger.debug(f"Test table {test_table.unique_name} already exists")

        return test_table

    def verify_test_table_row_count(self, test_table_specification: TestTableSpecification) -> bool:
        expected_row_values = test_table_specification.row_values
        if expected_row_values is None:
            expected_row_values = (
                []
            )  # This is a table that is not expected to have any rows. We should check that as well!
        row_count_sql = f"SELECT COUNT(*) FROM {self.data_source_impl.sql_dialect.qualify_dataset_name(self.dataset_prefix, test_table_specification.unique_name)}"
        row_count = self.data_source_impl.execute_query(row_count_sql)
        try:
            row_count_int = int(row_count.rows[0][0])
        except ValueError:
            row_count_int = None

        if row_count_int is not None and row_count_int != len(expected_row_values):
            logger.warning(
                f"Test table {test_table_specification.unique_name} has {row_count_int} rows, expected {len(expected_row_values)}"
            )
            logger.warning(f"Attempting to drop and recreate table {test_table_specification.unique_name}")
            return False
        return True

    def _create_test_table_python_object(self, test_table_specification: TestTableSpecification) -> TestTable:
        columns: list[TestColumn] = []
        for test_column_specification in test_table_specification.columns:
            data_source_sql_data_type: SqlDataType = (
                self.data_source_impl.sql_dialect.map_test_sql_data_type_to_data_source(
                    source_data_type=test_column_specification.sql_data_type
                )
            )
            test_column: TestColumn = TestColumn(
                name=test_column_specification.column_name,
                sql_data_type=data_source_sql_data_type,
            )
            columns.append(test_column)

        sql_dialect = self.data_source_impl.sql_dialect

        return TestTable(
            data_source_name=self.data_source_impl.name,
            dataset_prefix=self.dataset_prefix,
            code_name=test_table_specification.unique_name,
            unique_name=test_table_specification.unique_name,
            qualified_name=sql_dialect.qualify_dataset_name(
                dataset_prefix=self.dataset_prefix,
                dataset_name=test_table_specification.unique_name,
            ),
            columns=columns,
            row_values=test_table_specification.row_values,
            table_type=test_table_specification.table_type,
        )

    def _create_and_insert_test_table(self, test_table: TestTable) -> None:
        self._create_test_table(test_table)
        self._insert_test_table_rows(test_table)

    def _create_test_table(self, test_table: TestTable) -> None:
        my_create_table = CREATE_TABLE(
            fully_qualified_table_name=test_table.qualified_name,
            columns=[
                CREATE_TABLE_COLUMN(name=column.name, type=column.sql_data_type)
                for column in test_table.columns.values()
            ],
        )
        sql: str = self.data_source_impl.sql_dialect.build_create_table_sql(my_create_table)
        self.data_source_impl.execute_update(sql)

    def _insert_test_table_rows(self, test_table: TestTable) -> None:
        sql: str = self._insert_test_table_rows_sql(test_table)
        if sql:
            self.data_source_impl.execute_update(sql)

    def _insert_test_table_rows_sql(self, test_table: TestTable) -> str:
        if test_table.row_values:
            insert_into_sql = self.data_source_impl.sql_dialect.build_insert_into_sql(
                INSERT_INTO(
                    fully_qualified_table_name=test_table.qualified_name,
                    values=[VALUES_ROW(row) for row in test_table.row_values],
                    columns=[COLUMN(column.name) for column in test_table.columns.values()],
                )
            )
            return insert_into_sql

    def _drop_test_table(self, table_name: str) -> None:
        # Note, this is not a fully qualified table name, it's just the table name.
        # So we need to qualify it ourselves.
        fully_qualified_table_name = self.data_source_impl.sql_dialect.qualify_dataset_name(
            dataset_prefix=self.dataset_prefix,
            dataset_name=table_name,
        )
        my_drop_table = DROP_TABLE(fully_qualified_table_name=fully_qualified_table_name, cascade=True)
        sql: str = self.data_source_impl.sql_dialect.build_drop_table_sql(my_drop_table)
        self.data_source_impl.execute_update(sql)

    def query_existing_test_views(self) -> list[FullyQualifiedViewName]:
        metadata_tables_query: MetadataTablesQuery = self.data_source_impl.create_metadata_tables_query()
        fully_qualified_view_names: list[FullyQualifiedViewName] = metadata_tables_query.execute(
            database_name=self.extract_database_from_prefix(),
            schema_name=self.extract_schema_from_prefix(),
            include_table_name_like_filters=["SODATEST_%"],
            types_to_return=[TableType.VIEW],
        )
        return fully_qualified_view_names

    def query_existing_materialized_test_views(self) -> list[FullyQualifiedMaterializedViewName]:
        metadata_tables_query: MetadataTablesQuery = self.data_source_impl.create_metadata_tables_query()
        fully_qualified_view_names: list[FullyQualifiedMaterializedViewName] = metadata_tables_query.execute(
            database_name=self.extract_database_from_prefix(),
            schema_name=self.extract_schema_from_prefix(),
            include_table_name_like_filters=["SODATEST_%"],
            types_to_return=[TableType.MATERIALIZED_VIEW],
        )
        return fully_qualified_view_names

    def query_existing_test_view_names(self) -> list[str]:
        fully_qualified_view_names = self.query_existing_test_views()
        return [
            fully_qualified_view_name.view_name
            for fully_qualified_view_name in fully_qualified_view_names
            if isinstance(fully_qualified_view_name, FullyQualifiedViewName)
        ]

    def query_existing_test_materialized_view_names(self) -> list[str]:
        fully_qualified_view_names = self.query_existing_materialized_test_views()
        return [
            fully_qualified_view_name.materialized_view_name
            for fully_qualified_view_name in fully_qualified_view_names
            if isinstance(fully_qualified_view_name, FullyQualifiedMaterializedViewName)
        ]

    def create_view_from_test_table(self, test_table: TestTable) -> TestTable:
        view_name = f"{test_table.unique_name}_view"
        view_name = self.data_source_impl.sql_dialect.metadata_casify(view_name)
        existing_view_names = self.query_existing_test_view_names()
        if view_name in existing_view_names:
            self._drop_test_view(view_name)
        my_create_view = CREATE_VIEW(
            fully_qualified_view_name=self.data_source_impl.sql_dialect.qualify_dataset_name(
                dataset_prefix=self.dataset_prefix,
                dataset_name=view_name,
            ),
            select_elements=[
                SELECT(STAR()),
                FROM(self.data_source_impl.sql_dialect.get_from_name_from_qualified_name(test_table.qualified_name)),
            ],
        )
        sql: str = self.data_source_impl.sql_dialect.build_create_view_sql(my_create_view)
        self.data_source_impl.execute_update(sql)

        view_object = deepcopy(test_table)
        view_object.unique_name = view_name
        view_object.qualified_name = self.data_source_impl.sql_dialect.qualify_dataset_name(
            dataset_prefix=self.dataset_prefix,
            dataset_name=view_name,
        )

        return view_object

    def _drop_test_view(self, view_name: str) -> None:
        fully_qualified_view_name = self.data_source_impl.sql_dialect.qualify_dataset_name(
            dataset_prefix=self.dataset_prefix,
            dataset_name=view_name,
        )
        my_drop_view = DROP_VIEW(
            fully_qualified_view_name=fully_qualified_view_name,
        )
        sql: str = self.data_source_impl.sql_dialect.build_drop_view_sql(my_drop_view)
        self.data_source_impl.execute_update(sql)

    def create_materialized_view_from_test_table(self, test_table: TestTable) -> TestTable:
        view_name = f"{test_table.unique_name}_mv"
        view_name = self.data_source_impl.sql_dialect.metadata_casify(view_name)
        existing_view_names = self.query_existing_test_materialized_view_names()
        if view_name in existing_view_names:
            self._drop_test_materialized_view(view_name)
        my_create_view = CREATE_MATERIALIZED_VIEW(
            fully_qualified_view_name=self.data_source_impl.sql_dialect.qualify_dataset_name(
                dataset_prefix=self.dataset_prefix,
                dataset_name=view_name,
            ),
            select_elements=[
                SELECT(STAR()),
                FROM(self.data_source_impl.sql_dialect.get_from_name_from_qualified_name(test_table.qualified_name)),
            ],
        )
        sql: str = self.data_source_impl.sql_dialect.build_create_materialized_view_sql(my_create_view)
        self.data_source_impl.execute_update(sql)

        view_object = deepcopy(test_table)
        view_object.unique_name = view_name
        view_object.qualified_name = self.data_source_impl.sql_dialect.qualify_dataset_name(
            dataset_prefix=self.dataset_prefix,
            dataset_name=view_name,
        )
        return view_object

    def _drop_test_materialized_view(self, view_name: str) -> None:
        fully_qualified_view_name = self.data_source_impl.sql_dialect.qualify_dataset_name(
            dataset_prefix=self.dataset_prefix,
            dataset_name=view_name,
        )
        my_drop_view = DROP_MATERIALIZED_VIEW(
            fully_qualified_view_name=fully_qualified_view_name,
        )
        sql: str = self.data_source_impl.sql_dialect.build_drop_materialized_view_sql(my_drop_view)
        self.data_source_impl.execute_update(sql)

    def get_parse_errors_str(self, contract_yaml_str: str) -> str:
        contract_yaml_str: str = dedent(contract_yaml_str).strip()
        contract_verification_session_result = ContractVerificationSession.execute(
            contract_yaml_sources=[ContractYamlSource.from_str(contract_yaml_str)], only_validate_without_execute=True
        )
        return contract_verification_session_result.get_errors_str()

    def assert_contract_error(self, contract_yaml_str: str, variables: Optional[dict[str, str]] = None) -> str:
        contract_yaml_str: str = dedent(contract_yaml_str).strip()

        contract_verification_session_result: ContractVerificationSessionResult = ContractVerificationSession.execute(
            contract_yaml_sources=[
                ContractYamlSource.from_str(yaml_str=contract_yaml_str, file_path="yaml_string.yml")
            ],
            only_validate_without_execute=True,
            variables=variables,
            data_source_impls=[self.data_source_impl],
            soda_cloud_impl=self.soda_cloud,
            soda_cloud_use_agent=self.use_agent,
            soda_cloud_publish_results=True,
        )

        errors_str: str = contract_verification_session_result.get_errors_str()
        if not errors_str:
            raise AssertionError(f"Expected contract execution errors, but got none")
        return errors_str

    def assert_contract_pass(
        self,
        test_table: TestTable,
        contract_yaml_str: str,
        variables: Optional[dict[str, str]] = None,
        dwh_data_source_file_path: Optional[str] = None,
        extra_data_source_impls: list[DataSourceImpl] = [],
        check_paths: Optional[list[str]] = None,
        publish_results: bool = True,
    ) -> ContractVerificationResult:
        contract_verification_session_result: ContractVerificationSessionResult = self.verify_contract(
            contract_yaml_str=contract_yaml_str,
            test_table=test_table,
            variables=variables,
            dwh_data_source_file_path=dwh_data_source_file_path,
            extra_data_source_impls=extra_data_source_impls,
            check_paths=check_paths,
            publish_results=publish_results,
        )
        if not isinstance(contract_verification_session_result, ContractVerificationSessionResult):
            raise AssertionError(f"No contract verification result session")
        if not contract_verification_session_result.is_ok:
            raise AssertionError(f"Expected contract verification passed")
        if len(contract_verification_session_result.contract_verification_results) == 0:
            raise AssertionError(f"No contract verification results")

        # Final assert, is_ok does not guarantee all checks passed (could be skipped or non evaluated)
        assert contract_verification_session_result.is_passed, "Contract verification session result should be passed"

        return contract_verification_session_result.contract_verification_results[0]

    def assert_contract_ok(
        self,
        test_table: TestTable,
        contract_yaml_str: str,
        variables: Optional[dict[str, str]] = None,
        dwh_data_source_file_path: Optional[str] = None,
        extra_data_source_impls: list[DataSourceImpl] = [],
        check_paths: Optional[list[str]] = None,
        publish_results: bool = True,
    ) -> ContractVerificationResult:
        contract_verification_session_result: ContractVerificationSessionResult = self.verify_contract(
            contract_yaml_str=contract_yaml_str,
            test_table=test_table,
            variables=variables,
            dwh_data_source_file_path=dwh_data_source_file_path,
            extra_data_source_impls=extra_data_source_impls,
            check_paths=check_paths,
            publish_results=publish_results,
        )
        if not isinstance(contract_verification_session_result, ContractVerificationSessionResult):
            raise AssertionError(f"No contract verification result session")
        if not contract_verification_session_result.is_ok:
            raise AssertionError(f"Expected contract verification passed")
        if len(contract_verification_session_result.contract_verification_results) == 0:
            raise AssertionError(f"No contract verification results")

        return contract_verification_session_result.contract_verification_results[0]

    def assert_contract_fail(
        self,
        test_table: TestTable,
        contract_yaml_str: str,
        variables: Optional[dict[str, str]] = None,
        dwh_data_source_file_path: Optional[str] = None,
        extra_data_source_impls: list[DataSourceImpl] = [],
        publish_results: bool = True,
    ) -> ContractVerificationResult:
        contract_verification_session_result: ContractVerificationSessionResult = self.verify_contract(
            contract_yaml_str=contract_yaml_str,
            test_table=test_table,
            variables=variables,
            dwh_data_source_file_path=dwh_data_source_file_path,
            extra_data_source_impls=extra_data_source_impls,
            publish_results=publish_results,
        )
        if contract_verification_session_result.is_ok:
            raise AssertionError(f"Expected contract verification failed")
        return contract_verification_session_result.contract_verification_results[0]

    def assert_contract_warn(
        self,
        test_table: TestTable,
        contract_yaml_str: str,
        variables: Optional[dict[str, str]] = None,
        dwh_data_source_file_path: Optional[str] = None,
        extra_data_source_impls: list[DataSourceImpl] = [],
        publish_results: bool = True,
    ) -> ContractVerificationResult:
        contract_verification_session_result: ContractVerificationSessionResult = self.verify_contract(
            contract_yaml_str=contract_yaml_str,
            test_table=test_table,
            variables=variables,
            dwh_data_source_file_path=dwh_data_source_file_path,
            extra_data_source_impls=extra_data_source_impls,
            publish_results=publish_results,
        )
        if not contract_verification_session_result.is_warned:
            raise AssertionError(f"Expected contract verification warned")
        return contract_verification_session_result.contract_verification_results[0]

    def verify_contract(
        self,
        contract_yaml_str: str,
        test_table: Optional[TestTable] = None,
        variables: Optional[dict] = None,
        dwh_data_source_file_path: Optional[str] = None,
        extra_data_source_impls: list[DataSourceImpl] = [],
        check_paths: Optional[list[str]] = None,
        publish_results: bool = True,
    ) -> ContractVerificationSessionResult:
        contract_yaml_str = self._dedent_strip_and_prepend_dataset(contract_yaml_str, test_table)
        logger.debug(f"Contract:\n{contract_yaml_str}")

        return ContractVerificationSession.execute(
            contract_yaml_sources=[ContractYamlSource.from_str(contract_yaml_str)],
            variables=variables,
            data_source_impls=[self.data_source_impl, *extra_data_source_impls],
            soda_cloud_impl=self.soda_cloud,
            soda_cloud_use_agent=self.use_agent,
            soda_cloud_publish_results=publish_results,
            dwh_data_source_file_path=dwh_data_source_file_path,
            check_paths=check_paths,
        )

    def _dedent_strip_and_prepend_dataset(self, contract_yaml_str: str, test_table: Optional[TestTable]):
        checks_contract_yaml_str = dedent(contract_yaml_str).strip()
        if test_table:
            header_contract_yaml_str: str = f"dataset: {self.build_dqn(test_table)}\n"
            # This asserts that any "columns" statement in a check is single line.
            columns_contract_yaml_str: str = "columns: []\n" if not "columns:\n" in checks_contract_yaml_str else ""
            checks_contract_yaml_str = header_contract_yaml_str + columns_contract_yaml_str + checks_contract_yaml_str
        return checks_contract_yaml_str

    def build_dqn(self, test_table: TestTable) -> str:
        dqn_parts: list[str] = [self.data_source_impl.name] + self.dataset_prefix + [test_table.unique_name]
        dqn: str = "/".join(dqn_parts)
        return dqn

    def test_method_ended(self) -> None:
        # self.data_source_impl.data_source_connection.rollback() #TODO: this was originally done to theoretically speed up tests, but needs some datasource-specific work.
        self.soda_cloud = None
        self.use_agent = False

    def quote_column(self, column_name: str) -> str:
        """For shorter notation in the tests, we can just point it to the dialect."""
        return self.data_source_impl.sql_dialect.quote_column(column_name)

    def get_qualified_name_from_test_table(self, test_table: TestTable) -> str:
        return self.data_source_impl.sql_dialect.qualify_dataset_name(
            dataset_prefix=self.dataset_prefix,
            dataset_name=test_table.unique_name,
        )

    @classmethod
    def create_test_table_with_all_datatypes(cls, table_purpose: str, number_of_rows: int) -> TestTableSpecification:
        """
        Creates a test table with all data types.
        The smallint_default column will have a unique value for each row.
        """
        test_rows = [
            (
                # Chars
                "a",
                "abc",
                # Varchars
                "a",
                "abc",
                # Texts
                "a",
                # Integers
                i,
                100,
                100000,
                # Decimals
                1.0,
                100.0,
                100000.1234567890,
                # Numerics
                1.0,
                100.0,
                100000.1234567890,
                # Floats
                1.1234,
                # Doubles
                1.1234567890,
                # Timestamps
                datetime.datetime(2021, 1, 1, 1, 1, 1),
                datetime.datetime(2022, 1, 1, 1, 1, 1, 12),
                # Timestamp Zones
                datetime.datetime(2023, 1, 1, 1, 1, 1),
                datetime.datetime(2024, 1, 1, 1, 1, 1, 12),
                # Dates
                datetime.date(2025, 1, 1),
                # Times
                datetime.time(1, 2, 3),
                # Booleans
                True,
            )
            for i in range(number_of_rows)
        ]
        test_table_specification = (
            TestTableSpecification.builder()
            .table_purpose(table_purpose)
            .column_char("char_default")
            .column_char("char_w_length", 100)
            .column_varchar("varchar_default")
            .column_varchar("varchar_w_length", 100)
            .column_text("text_default")
            .column_smallint("smallint_default")
            .column_integer("integer_default")
            .column_bigint("bigint_default")
            .column_decimal("decimal_default")
            .column_decimal("decimal_w_precision", 10)
            .column_decimal("decimal_w_precision_and_scale", 10, 2)
            .column_numeric("numeric_default")
            .column_numeric("numeric_w_precision", 10)
            .column_numeric("numeric_w_precision_and_scale", 10, 2)
            .column_float("float_default")
            .column_double("double_default")
            .column_timestamp("timestamp_default")
            .column_timestamp("timestamp_w_precision", 2)
            .column_timestamp_tz("timestamp_tz_default")
            .column_timestamp_tz("timestamp_tz_w_precision", 4)
            .column_date("date_default")
            .column_time("time_default")
            .column_boolean("boolean_default")
            .rows(rows=test_rows)
            .build()
        )
        return test_table_specification

    def get_column_mappings(self) -> dict[str, SodaDataTypeName]:
        return {
            "char_default": SodaDataTypeName.CHAR,
            "char_w_length": SodaDataTypeName.CHAR,
            "varchar_default": SodaDataTypeName.VARCHAR,
            "varchar_w_length": SodaDataTypeName.VARCHAR,
            "text_default": SodaDataTypeName.TEXT,
            "smallint_default": SodaDataTypeName.SMALLINT,
            "integer_default": SodaDataTypeName.INTEGER,
            "bigint_default": SodaDataTypeName.BIGINT,
            "decimal_default": SodaDataTypeName.DECIMAL,
            "decimal_w_precision": SodaDataTypeName.DECIMAL,
            "decimal_w_precision_and_scale": SodaDataTypeName.DECIMAL,
            "numeric_default": SodaDataTypeName.NUMERIC,
            "numeric_w_precision": SodaDataTypeName.NUMERIC,
            "numeric_w_precision_and_scale": SodaDataTypeName.NUMERIC,
            "float_default": SodaDataTypeName.FLOAT,
            "double_default": SodaDataTypeName.DOUBLE,
            "timestamp_default": SodaDataTypeName.TIMESTAMP,
            "timestamp_w_precision": SodaDataTypeName.TIMESTAMP,
            "timestamp_tz_default": SodaDataTypeName.TIMESTAMP_TZ,
            "timestamp_tz_w_precision": SodaDataTypeName.TIMESTAMP_TZ,
            "date_default": SodaDataTypeName.DATE,
            "time_default": SodaDataTypeName.TIME,
            "boolean_default": SodaDataTypeName.BOOLEAN,
        }

    def map_table_type_to_short_string(self, table_type: TableType) -> str:
        return {
            TableType.TABLE: "t",
            TableType.MATERIALIZED_VIEW: "mv",
            TableType.VIEW: "v",
        }[table_type]



from __future__ import annotations

import json
import logging
from dataclasses import dataclass
from enum import Enum
from io import BytesIO
from tempfile import TemporaryFile
from typing import Optional

from requests import Response
from soda_core.common.dataset_identifier import DatasetIdentifier
from soda_core.common.soda_cloud import SodaCloud
from soda_core.common.soda_cloud_dto import DatasetConfigurationDTO


class MockHttpMethod(Enum):
    POST = "post"
    GET = "get"


class MockResponse(Response):
    def __init__(
        self,
        method: MockHttpMethod = MockHttpMethod.POST,
        status_code: int = 200,
        headers: Optional[dict[str, str]] = None,
        json_object: Optional[dict] = None,
    ):
        super().__init__()
        self.method: MockHttpMethod = method
        self.status_code = status_code
        if isinstance(headers, dict):
            self.headers.update(headers)
        # The json_object is stored for overwriting later. See below in _resolve_check_identities
        self.json_object: Optional[dict] = None
        self.set_json_object(json_object)

    def set_json_object(self, json_object: Optional[dict]):
        self.json_object = json_object
        rows_json_str = json.dumps(json_object)
        rows_json_bytes = bytearray(rows_json_str, "utf-8")
        self.raw = BytesIO(rows_json_bytes)


@dataclass
class MockRequest:
    request_log_name: str = (None,)
    url: Optional[str] = (None,)
    headers: dict[str, str] = (None,)
    json: Optional[dict] = (None,)
    data: Optional[TemporaryFile] = None


class MockSodaCloud(SodaCloud):
    def __init__(self, responses: Optional[list[Optional[MockResponse]]] = None):
        super().__init__(
            host="mock.soda.io",
            api_key_id="mock-key-id",
            api_key_secret="mock-key-secret",
            token="mock-token",
            port="9999",
            scheme="https",
        )
        self.requests: list[MockRequest] = []
        self.responses: list[Optional[MockResponse]] = responses if isinstance(responses, list) else []
        self.dataset_configurations: dict[DatasetIdentifier, DatasetConfigurationDTO] = {}

    def _http_post(
        self,
        request_log_name: str = None,
        url: Optional[str] = None,
        headers: dict[str, str] = None,
        json: Optional[dict] = None,
        data: Optional[TemporaryFile] = None,
    ) -> Response:
        return self._http_handle(method=MockHttpMethod.POST, url=url, headers=headers, json=json, data=data)

    def _http_get(
        self,
        url: Optional[str] = None,
        headers: dict[str, str] = None,
        json: Optional[dict] = None,
        data: Optional[TemporaryFile] = None,
    ) -> Response:
        return self._http_handle(method=MockHttpMethod.GET, url=url, headers=headers, json=json, data=data)

    def _http_handle(
        self,
        method: MockHttpMethod,
        url: Optional[str],
        headers: dict[str, str],
        json: Optional[dict],
        data: Optional[TemporaryFile],
    ) -> Response:
        self.requests.append(MockRequest(url=url, headers=headers, json=json, data=data))
        if self.responses:
            response = self.responses.pop(0)
            if isinstance(response, MockResponse):
                if method != response.method:
                    raise AssertionError("Wrong response method")
                if self._is_send_scan_results_request(json):
                    self._resolve_check_identities(json, response)
                logging.debug(f"MockSodaCloud responds to {method} {url} with provided response")
                return response
        logging.debug(f"MockSodaCloud responds to {method} {url} with default empty 200 OK response")
        return MockResponse(status_code=200, headers={}, json_object={})

    def _is_send_scan_results_request(self, request_json: Optional[dict]) -> bool:
        return (
            isinstance(request_json, dict)
            and "type" in request_json
            and request_json["type"] == "sodaCoreInsertScanResults"
        )

    def _resolve_check_identities(self, request_json: dict, response: MockResponse):
        request_check_identities: list[str] = []
        request_check_results: Optional[list[dict]] = request_json.get("checks")
        if request_check_results:
            for request_check_result in request_check_results:
                check_identities: Optional[dict] = request_check_result.get("identities")
                if isinstance(check_identities, dict) and "vc1" in check_identities:
                    request_check_identities.append(check_identities["vc1"])

        new_json_object: dict = response.json_object.copy()
        response_checks: Optional[list[dict]] = new_json_object.get("checks")
        if isinstance(response_checks, list) and len(response_checks) == len(request_check_identities):
            for index in range(0, len(response_checks)):
                response_check: dict = response_checks[index]
                response_check["identities"] = [request_check_identities[index]]

        response.set_json_object(new_json_object)

    def set_dataset_configuration_response(
        self, dataset_identifier: "DatasetIdentifier", dataset_configuration_dto: "DatasetConfigurationDTO"
    ):
        self.dataset_configurations[str(dataset_identifier)] = dataset_configuration_dto

    def fetch_dataset_configuration(
        self, dataset_identifier: "DatasetIdentifier"
    ) -> Optional["DatasetConfigurationDTO"]:
        if str(dataset_identifier) in self.dataset_configurations:
            return self.dataset_configurations[str(dataset_identifier)]

        # Return an empty DatasetConfigurationDTO for testing purposes. This is sufficient for tests that do not depend on specific configuration values.
        # This overrides the mock response setup in tests, but it would be annoying to have to set up the mock response for every test just to return an empty configuration.
        return DatasetConfigurationDTO()



from textwrap import dedent
from typing import Optional

from soda_core.contracts.contract_verification import CheckResult


def dedent_and_strip(text: str) -> str:
    return dedent(text).strip()


def get_diagnostic_value(check_result: CheckResult, diagnostic_name: str) -> Optional[any]:
    return check_result.diagnostic_metric_values.get(diagnostic_name)



#!/usr/bin/env python

import sys

from setuptools import setup

if sys.version_info < (3, 9):
    print("Error: soda-core requires at least Python 3.9")
    print("Error: Please upgrade your Python version to 3.9 or later")
    sys.exit(1)

package_name = "soda-tests"
package_version = "4.0.5b2"
description = "Soda Core V4 Tests"

requires = [f"soda-core=={package_version}", "pytz", "python-dateutil"]

setup(
    name=package_name,
    version=package_version,
    install_requires=requires,
    description="Soda Core Tests",
    package_dir={"": "src"},
    package_data={"": ["**/*.json"]},
    include_package_data=True,
    python_requires=">=3.9",
)



from dataclasses import dataclass
from typing import Any, Optional

import pytest
from soda_core.common.data_source_impl import DataSourceImpl
from soda_core.common.logs import Logs
from soda_core.common.yaml import DataSourceYamlSource


@dataclass
class TestConnection:
    """
    Define a connection YAML string and test that it can be parsed, used to connect, and run queries.

    You can provide invalid connection parameters and test that an expected error is raised.  See
    example implementation in soda-bigquery/tests/data_sources/test_bigquery.py
    """

    test_name: str
    connection_yaml_str: str
    valid_yaml: Optional[bool] = True
    valid_connection_params: Optional[bool] = True
    query_should_succeed: Optional[bool] = True
    expected_yaml_error: Optional[str] = None
    expected_connection_error: Optional[str] = None
    expected_query_error: Optional[str] = None
    monkeypatches: Optional[dict[str, Any]] = None

    def create_data_source_yaml(self) -> DataSourceYamlSource:
        return DataSourceYamlSource.from_str(yaml_str=self.connection_yaml_str)

    def create_data_source_impl(self, data_source_yaml_source: DataSourceYamlSource) -> DataSourceImpl:
        return DataSourceImpl.from_yaml_source(data_source_yaml_source)

    def test(self, monkeypatch: Optional[pytest.MonkeyPatch] = None):
        if self.monkeypatches:
            for module, mock in self.monkeypatches.items():
                monkeypatch.setattr(module, mock)

        logs = Logs()
        data_source_yaml = self.create_data_source_yaml()
        data_source_impl = None

        try:
            if self.valid_yaml:
                data_source_impl = self.create_data_source_impl(data_source_yaml)
                assert not logs.has_errors
            else:
                with pytest.raises(Exception) as exc_info:
                    self.create_data_source_impl(data_source_yaml)

                assert self.expected_yaml_error in str(exc_info.value)
                # do not try to connect or query if YAML parsing failed
                return

            if self.valid_connection_params:
                data_source_impl.open_connection()
                if logs.has_errors:
                    error_msg = "Connection failed unexpectedly with error: " + logs.get_errors_str()
                    raise RuntimeError(error_msg)
            else:
                data_source_impl.open_connection()
                assert logs.has_errors
                assert self.expected_connection_error in logs.get_errors_str()
                # do not try to query if connection failed
                return

            if self.query_should_succeed:
                data_source_impl.execute_query("SELECT 1")
                if logs.has_errors:
                    error_msg = "Query failed unexpectedly with error: " + logs.get_errors_str()
                    raise RuntimeError(error_msg)
            else:
                with pytest.raises(Exception) as exc_info:
                    data_source_impl.execute_query("SELECT 1")
                assert self.expected_query_error in str(exc_info.value)
                return
        finally:
            if data_source_impl:
                data_source_impl.close_connection()  # safe even if no connection open



from __future__ import annotations

from dataclasses import dataclass
from typing import Iterable, Optional

from soda_core.common.consistent_hash_builder import ConsistentHashBuilder
from soda_core.common.dataset_identifier import DatasetIdentifier
from soda_core.common.metadata_types import (
    ColumnMetadata,
    SodaDataTypeName,
    SqlDataType,
)
from soda_core.common.statements.table_types import TableType

MAX_TABLE_NAME_LENGTH = 63  # Postgres max table name length.  Needed or else table names may be incorrectly truncated.


# This should technically be pulled from SQL dialect and vary by data source, but we don't have access to that here
# So far every data source can handle 63 characters.
class TestTableSpecificationBuilder:
    """
    See TestTableSpecification for documentation
    """

    __test__ = False

    # All test table names to verify that every test table name is used only once in the whole test suite.
    __names = []

    def __init__(self):
        self._table_purpose: Optional[str] = None
        self._columns: list[ColumnMetadata] = []
        self._rows: Optional[list[tuple]] = None
        self._table_type: TableType = TableType.TABLE

    def table_purpose(self, table_purpose: str) -> TestTableSpecificationBuilder:
        """
        The middle section of the table name indicating the purpose.
        The full table name will be composed of
          * The "SODATEST_" prefix
          * The test table name
          * A hash of the table specification (for recreating tables only when the schema or data changes)
        """
        # Some DB engines may fail with test table names that are too long
        # we could try intelligently truncate and hash the table names but simpler to just enforce the limit
        max_table_purpose_length = MAX_TABLE_NAME_LENGTH - len("SODATEST_") - len("_12345678")
        if len(table_purpose) > max_table_purpose_length:
            raise ValueError(
                f"Max table purpose length is {max_table_purpose_length} characters.  Table purpose {table_purpose} is {len(table_purpose)} characters long."
            )
        self._table_purpose = table_purpose
        return self

    def column(
        self,
        name: str,
        soda_data_type_name: SodaDataTypeName,
        character_maximum_length: Optional[int] = None,
        numeric_precision: Optional[int] = None,
        numeric_scale: Optional[int] = None,
        datetime_precision: Optional[int] = None,
    ) -> TestTableSpecificationBuilder:
        self._columns.append(
            ColumnMetadata(
                column_name=name,
                sql_data_type=SqlDataType(
                    name=soda_data_type_name,
                    character_maximum_length=character_maximum_length,
                    numeric_precision=numeric_precision,
                    numeric_scale=numeric_scale,
                    datetime_precision=datetime_precision,
                ),
            )
        )
        return self

    def column_char(
        self,
        name: str,
        character_maximum_length: Optional[int] = None,
    ) -> TestTableSpecificationBuilder:
        return self.column(
            name=name, soda_data_type_name=SodaDataTypeName.CHAR, character_maximum_length=character_maximum_length
        )

    def column_varchar(
        self,
        name: str,
        character_maximum_length: Optional[int] = None,
    ) -> TestTableSpecificationBuilder:
        return self.column(
            name=name,
            soda_data_type_name=SodaDataTypeName.VARCHAR,
            character_maximum_length=character_maximum_length,
        )

    def column_text(self, name: str) -> TestTableSpecificationBuilder:
        return self.column(name=name, soda_data_type_name=SodaDataTypeName.TEXT)

    def column_smallint(self, name) -> TestTableSpecificationBuilder:
        return self.column(name=name, soda_data_type_name=SodaDataTypeName.SMALLINT)

    def column_integer(self, name) -> TestTableSpecificationBuilder:
        return self.column(name=name, soda_data_type_name=SodaDataTypeName.INTEGER)

    def column_bigint(self, name) -> TestTableSpecificationBuilder:
        return self.column(name=name, soda_data_type_name=SodaDataTypeName.BIGINT)

    def column_numeric(
        self,
        name: str,
        numeric_precision: Optional[int] = None,
        numeric_scale: Optional[int] = None,
    ) -> TestTableSpecificationBuilder:
        return self.column(
            name=name,
            soda_data_type_name=SodaDataTypeName.NUMERIC,
            numeric_precision=numeric_precision,
            numeric_scale=numeric_scale,
        )

    def column_decimal(
        self,
        name: str,
        numeric_precision: Optional[int] = None,
        numeric_scale: Optional[int] = None,
    ) -> TestTableSpecificationBuilder:
        return self.column(
            name=name,
            soda_data_type_name=SodaDataTypeName.DECIMAL,
            numeric_precision=numeric_precision,
            numeric_scale=numeric_scale,
        )

    def column_float(self, name) -> TestTableSpecificationBuilder:
        return self.column(name=name, soda_data_type_name=SodaDataTypeName.FLOAT)

    def column_double(self, name) -> TestTableSpecificationBuilder:
        return self.column(name=name, soda_data_type_name=SodaDataTypeName.DOUBLE)

    def column_date(self, name) -> TestTableSpecificationBuilder:
        return self.column(name=name, soda_data_type_name=SodaDataTypeName.DATE)

    def column_time(self, name) -> TestTableSpecificationBuilder:
        return self.column(name=name, soda_data_type_name=SodaDataTypeName.TIME)

    def column_timestamp(
        self,
        name: str,
        datetime_precision: Optional[int] = None,
    ) -> TestTableSpecificationBuilder:
        return self.column(
            name=name,
            soda_data_type_name=SodaDataTypeName.TIMESTAMP,
            datetime_precision=datetime_precision,
        )

    def column_timestamp_tz(
        self,
        name: str,
        datetime_precision: Optional[int] = None,
    ) -> TestTableSpecificationBuilder:
        return self.column(
            name=name,
            soda_data_type_name=SodaDataTypeName.TIMESTAMP_TZ,
            datetime_precision=datetime_precision,
        )

    def column_boolean(self, name) -> TestTableSpecificationBuilder:
        return self.column(name=name, soda_data_type_name=SodaDataTypeName.BOOLEAN)

    def rows(self, rows: list[tuple]) -> TestTableSpecificationBuilder:
        """
        Python objects that will be translated to data source specific literals
        Usage:
            .rows([
                ('1', 'US'),
                ('2', 'US'),
                ('3', 'BE')
            ])
        """
        self._rows = rows
        return self

    def build(self) -> TestTableSpecification:
        assert isinstance(self._table_purpose, str)
        # Ensure unique table data names
        name_lower = self._table_purpose.lower()
        if name_lower in self.__names:
            raise AssertionError(
                f"Duplicate test table purpose detected: {self._table_purpose}.  In the codebase, the table_purpose "
                f'of every test table should be unique.  Search for .table_purpose("{self._table_purpose}") and you '
                f"should find multiple places in the test codebase where the same table_purpose is created."
            )
        self.__names.append(name_lower)
        unique_name = f"SODATEST_{self._table_purpose}_{self.__test_table_hash()}"
        return TestTableSpecification(
            name=self._table_purpose,
            columns=self._columns,
            row_values=self._rows,
            unique_name=unique_name,
            table_type=self._table_type,
        )

    def __test_table_hash(self) -> str:
        consistent_hash_builder: ConsistentHashBuilder = ConsistentHashBuilder()
        consistent_hash_builder.add(self._table_purpose)
        consistent_hash_builder.add("columns")
        for test_column in self._columns:
            consistent_hash_builder.add(test_column.column_name)
            consistent_hash_builder.add(test_column.sql_data_type.name)
            if test_column.sql_data_type:
                consistent_hash_builder.add(test_column.sql_data_type.character_maximum_length)
                consistent_hash_builder.add(test_column.sql_data_type.numeric_precision)
                consistent_hash_builder.add(test_column.sql_data_type.numeric_scale)
                consistent_hash_builder.add(test_column.sql_data_type.datetime_precision)
        if isinstance(self._rows, Iterable):
            for row in self._rows:
                consistent_hash_builder.add("row")
                if isinstance(row, Iterable):
                    for value in row:
                        consistent_hash_builder.add(value)
        return consistent_hash_builder.get_hash()


@dataclass
class TestTableSpecification:
    """
    Data model for specifying a test tables that are lazy created and initialized in the data source before they can
    be used in tests. It includes all information from which a concrete test table can be created, including the
    records.

    To minimize runtime overhead the test infrastructure recycles tables as long as the test table specification
    stay the same in the codebase.  To accomplish this test table names are composed of 3 parts:
          * The "SODATEST_" prefix
          * The table purpose
          * A hash of the table specification

    If a test table doesn't exist, it is created as part of the ensure_test_table method.  Tables remain
    in the data source after the tests.

    If the data or schema changes in the code of a test table specification, then the hash of the table specification
    will change.  The new table SODATEST_{table_purpose}_{new_table_hash} will be created and the
    old table SODATEST_{table_purpose}_{old_table_hash} will be deleted.

    Multiple tests can leverage the same test table specification.  The convention is that all the tests in one
    test file use 1 test table specification.  That way, reusing the same table for multiple tests is still supported
    and the impact of changing a test table specification is limited to the file you're working in.
    """

    __test__ = False

    @staticmethod
    def builder() -> TestTableSpecificationBuilder:
        return TestTableSpecificationBuilder()

    name: Optional[str]
    columns: list[ColumnMetadata]
    row_values: Optional[list[tuple]]
    unique_name: Optional[str]
    table_type: TableType


class TestTable:
    __test__ = False

    def __init__(
        self,
        data_source_name: str,
        dataset_prefix: list[str],
        code_name: str,
        unique_name: str,
        qualified_name: str,
        columns: list[TestColumn],
        row_values: Optional[list[tuple]],
        table_type: TableType = TableType.TABLE,
    ):
        self.data_source_name: str = data_source_name
        self.dataset_prefix: list[str] = dataset_prefix
        # Name of the test table in the code.
        self.code_name: str = code_name
        # Name that includes the hash of the contents of the table
        self.unique_name: str = unique_name
        # Fully qualified SQL name of the table
        self.qualified_name: str = qualified_name
        self.columns: dict[str, TestColumn] = {column.name: column for column in columns}
        self.row_values: Optional[list[tuple]] = row_values
        self.table_type: TableType = table_type

    def data_type(self, column_name: str) -> str:
        return self.columns[column_name].sql_data_type.name

    def get_dataset_qualified_name(self) -> str:
        slash_separated_prefixes: str = "/".join(self.dataset_prefix)
        return f"{self.data_source_name}/{slash_separated_prefixes}/{self.unique_name}"

    @property
    def dataset_identifier(self) -> "DatasetIdentifier":
        return DatasetIdentifier(
            data_source_name=self.data_source_name,
            prefixes=self.dataset_prefix,
            dataset_name=self.unique_name,
        )


class TestColumn:
    __test__ = False

    def __init__(self, name: str, sql_data_type: SqlDataType):
        self.name: str = name
        self.sql_data_type: SqlDataType = sql_data_type
